﻿#拡張子を*.pyに変えてコマンドラインからpython ファイル名.py で実行出来ます


import tkinter as tk
from tkinter import filedialog
import tkinter.messagebox as mbox
import json
from PIL import Image, ImageGrab, PngImagePlugin
import io
import math
import pickle
import random

def rotate(x, y, rad_angle):
    # 回転行列を計算
    cos_a = math.cos(rad_angle)
    sin_a = math.sin(rad_angle)
    # 回転後の座標を計算
    x_rot = x * cos_a - y * sin_a
    y_rot = x * sin_a + y * cos_a
    return x_rot, y_rot

def rotate_point(x, y, cx, cy, angle):
    # 角度をラジアンに変換
    theta = math.radians(angle)
    # 回転前の座標を中心座標からの相対位置に変換
    rel_x = x - cx
    rel_y = y - cy

    new_rel_x, new_rel_y = rotate(rel_x, rel_y, theta)
    # 回転後の座標を計算して返す
    new_x = new_rel_x + cx
    new_y = new_rel_y + cy
    return new_x, new_y

def ellipse_points(x, y, a, b, rad_angle, n, start_i=0, end_i=None):
    if not end_i:
        end_i = n
    # 多角形の頂点を計算
    points = []
    for i in range(start_i,end_i):
        theta = math.radians(360 * i / n)
        px, py = rotate(a * math.cos(theta), b * math.sin(theta), rad_angle)
        px += x
        py += y
        points.append(px)
        points.append(py)
    
    # 多角形を描画
    return points

def ovalline_points(x1,y1, x2,y2, width=10):
    cx = (x1+x2)/2
    cy = (y1+y2)/2
    l = ((x2-x1)**2 + (y2-y1)**2)**0.5
    rad_angle = math.atan2(y2-y1,x2-x1)
    points = ellipse_points(cx,cy,l/2,10/2,rad_angle,8)
    return points

def ovalline_points(x1,y1, x2,y2, width=10):
    # 楕円の割合
    alpha = 0.5

    n = 8

    cx = (x1+x2)/2
    cy = (y1+y2)/2
    
    nx, ny = rotate(1.0, 0.0, math.atan2(y2-y1,x2-x1) + math.radians(90))
    nx = nx * width*(1.0-alpha)/4
    ny = ny * width*(1.0-alpha)/4

    l = ((x2-x1)**2 + (y2-y1)**2)**0.5
    rad_angle = math.atan2(y2-y1,x2-x1)
    points1 = ellipse_points(cx+nx,cy+ny,l/2,10/2,rad_angle,n,0+1,n//2)
    points2 = ellipse_points(cx-nx,cy-ny,l/2,10/2,rad_angle,n,n//2+1,n)
    return [x1,y1] + [x1-nx,y1-ny] + points2 + [x2-nx,y2-ny] + [x2,y2] + [x2+nx,y2+ny] + points1 + [x1+nx,y1+ny]

def create_ovalline(canvas, x1,y1, x2,y2, color):
    points = ovalline_points(x1,y1,x2,y2)
    return canvas.create_polygon(*points, fill=color)

def coords_ovalline(canvas, obj, x1,y1, x2,y2):
    points = ovalline_points(x1,y1,x2,y2)
    return canvas.coords(obj, *points)

def endpoint_coordinates(canvas, obj):
    points = canvas.coords(obj)
    endpoint_i = len(points)//2
    x1,y1 = points[0:2]
    x2,y2 = points[endpoint_i:endpoint_i+2]
    return [x1,y1,x2,y2]

class EdgeManager:
    def __init__(self, canvas):
        self.canvas = canvas

#    def create_line(self,x1,y1,x2,y2):
        
#    def set_coords(self,obj,x1,y1,x2,y2):
        
#    def get_coords(self,obj):

class LineEdgeManager(EdgeManager):
    def __init__(self, canvas):
        super().__init__(canvas)

    def create_line(self,x1,y1,x2,y2,color):
        return self.canvas.create_line(x1, y1, x2, y2, width=10, fill=color)
        
    def set_coords(self,obj,x1,y1,x2,y2):
        self.canvas.coords(obj, x1, y1, x2, y2)

    def get_coords(self,obj):
        return self.canvas.coords(obj)

class EllipseEdgeManager(EdgeManager):
    def __init__(self, canvas):
        super().__init__(canvas)

    def create_line(self,x1,y1,x2,y2,color):
        return create_ovalline(self.canvas, x1,y1,x2,y2, color)

    def set_coords(self,obj,x1,y1,x2,y2):
        coords_ovalline(self.canvas, obj, x1, y1, x2, y2)

    def get_coords(self,obj):
        return endpoint_coordinates(self.canvas,obj)



NODE_RADIUS=5
class GraphVisualizer:
    def __init__(self, master):
        self.master = master
        self.graph = Graph()
        self.canvas = tk.Canvas(self.master, width=640, height=896, background="black")
        self.canvas.pack()
        self.canvas.bind("<MouseWheel>", self.zoom)
        self.canvas.bind("<Control-MouseWheel>", self.rotate)
        self.canvas.bind("<Motion>", self.on_motion)
        self.canvas.bind("<Leave>", self.on_leave_canvas)
        self.canvas.bind("<Double-Button-1>", self.show_submenu)
        self.canvas.bind("<Button-3>", self.on_button3_pressed)
        self.canvas.bind("<Button-1>", self.on_button3_pressed)

        self.edgeManagers = [EllipseEdgeManager(self.canvas), LineEdgeManager(self.canvas)]
        self.edgeManagerIndex = 0
        self.edgeManager = self.edgeManagers[self.edgeManagerIndex]

        # ファイルへの保存と読み込みのためのメニューバーを追加する
        menubar = tk.Menu(self.master)
        self.master.config(menu=menubar)

        filemenu = tk.Menu(menubar, tearoff=0)
        menubar.add_cascade(label="File", menu=filemenu)
        filemenu.add_command(label="Save as Image", command=self.save_as_image)
        filemenu.add_command(label="Load Pose (*png with embedded pose, or *.json)", command=lambda: self.load_file(False))


        # メニューバーを作成
        self.master.config(menu=menubar)

        # 「オプション」メニューを作成
        edit_menu = tk.Menu(menubar,tearoff=0)
        menubar.add_cascade(label="編集", menu=edit_menu)
        edit_menu.add_command(label="人を追加", command=self.add_human_graph)
        edit_menu.add_command(label="ファイルから人を追加", command= lambda : self.load_file(True))
#        edit_menu.add_command(label="ポーズデータが埋め込まれた画像ファイルから人を追加", command= lambda : self.load_pose_from_image(True))

        # 「オプション」メニューを作成
        option_menu = tk.Menu(menubar,tearoff=0)
        menubar.add_cascade(label="オプション", menu=option_menu)

        # 「サイズ変更」アイテムを作成
        option_menu.add_command(label="キャンバスサイズ変更", command=self.show_option_window)
        option_menu.add_command(label="ウィンドウの透明/半透明を切り替え", command=self.toggle_transparency)
        option_menu.add_command(label="線のスタイルを切り替え", command=self.toggle_line_style)

    def toggle_transparency(self):
        alpha = self.master.attributes("-alpha")
        if alpha == 1.0:
            self.master.attributes("-alpha", 0.5)
        else:
            self.master.attributes("-alpha", 1.0)

    def toggle_line_style(self):
        self.edgeManagerIndex = (self.edgeManagerIndex + 1) % len(self.edgeManagers)
        self.edgeManager = self.edgeManagers[self.edgeManagerIndex]
        self.draw_objects()

    def set_radius(self,item,radius):
        # ovalの座標を取得する
        x0, y0, x1, y1 = self.canvas.bbox(item)

        # 半径がradiusの円にする
        cx = (x0+x1)/2
        cy = (y0+y1)/2
        self.canvas.coords(item, cx-radius, cy-radius, cx+radius, cy+radius)

    def on_leave_canvas(self,event):
        for item in self.canvas.find_withtag("node"):
            self.set_radius(item,NODE_RADIUS)

    def on_motion(self,event):
        if self.graph.nodes:
            x, y = self.canvas.canvasx(event.x), self.canvas.canvasy(event.y)
            nearest_item = self.nodes[self.graph.find_nearest_node(x,y)]
            for item in self.canvas.find_withtag("node"):
                bbox = self.canvas.bbox(item)
                oval_x, oval_y = (bbox[0]+bbox[2])/2, (bbox[1]+bbox[3])/2
                distance = ((x - oval_x)**2 + (y - oval_y)**2)**0.5  # カーソルとovalの距離を計算する
                if item==nearest_item:  # 最も近いノードの場合
                    self.canvas.addtag_withtag("highlight", item)  # タグを追加してハイライト表示する
                    self.set_radius(item,NODE_RADIUS*2)
                else:
                    self.canvas.dtag(item, "highlight")  # タグを削除してハイライト表示を解除する
                    self.set_radius(item,NODE_RADIUS)

    def show_option_window(self):
        # オプションウィンドウを作成
        self.option_window = tk.Toplevel(self.master)
        self.option_window.title("サイズ変更")

        # 幅の入力フィールドを作成
        tk.Label(self.option_window, text="幅").grid(row=0, column=0)
        self.width_entry = tk.Entry(self.option_window)
        self.width_entry.grid(row=0, column=1)
        self.width_entry.insert(0, str(self.canvas["width"]))

        # 高さの入力フィールドを作成
        tk.Label(self.option_window, text="高さ").grid(row=1, column=0)
        self.height_entry = tk.Entry(self.option_window)
        self.height_entry.grid(row=1, column=1)
        self.height_entry.insert(0, str(self.canvas["height"]))

        # OKボタンを作成
        tk.Button(self.option_window, text="OK", command=self.update_canvas_size).grid(row=2, column=0, columnspan=2)

    def update_canvas_size(self):
        # 入力された幅と高さを取得し、Canvasのサイズを変更
        width = int(self.width_entry.get())
        height = int(self.height_entry.get())
        self.canvas.config(width=width, height=height)

        # オプションウィンドウを破棄
        self.option_window.destroy()

    def get_body_nodes(self,node):
        for start_i in range(len(self.graph.nodes)//18):
            body_nodes = self.graph.nodes[start_i*18:start_i*18+18]
            if node in body_nodes:
                return body_nodes

    def show_submenu(self,event):
        x, y = self.canvas.canvasx(event.x), self.canvas.canvasy(event.y)
        
        nodes = self.get_body_nodes(self.graph.find_nearest_node(x,y))


        # サブメニューを作成する
        submenu = tk.Menu(self.canvas, tearoff=0)
        submenu.add_command(label="コピー", command = lambda nodes=nodes : self.copy_human(nodes))
        submenu.add_command(label="左右反転", command = lambda nodes=nodes : self.flip_human(nodes))
        submenu.add_command(label="削除", command = lambda nodes=nodes : self.remove_human(nodes))
    
        # サブメニューを表示する
        submenu.post(event.x_root, event.y_root)

    def copy_human(self,nodes):

        to_new_node = {}
        for node in nodes:
            x,y = node.position
            color = node.color
            new_node = self.graph.add_node( (x+30,y+30), color )
            to_new_node[node] = new_node

        for edge in self.graph.edges_from_nodes(nodes):
            self.graph.add_edge( to_new_node[edge.start_node], to_new_node[edge.end_node], edge.color )

        self.draw_objects()

        return

    def flip_human(self,nodes):
        left_x =  min([node.position[0] for node in nodes])
        right_x =  max([node.position[0] for node in nodes])
        middle_x = (right_x+left_x)/2
        for node in nodes:
            x = node.position[0]
            node.position = (x -2*(x-middle_x), node.position[1] )
        self.update_canvas_objects()

    def remove_human(self,nodes):
        if mbox.askyesno("確認", "本当にこの人を削除してよろしいですか？"):
            for node in nodes:
                for edge in node.edges:
                    try:
                        del self.edges[edge]
                    except KeyError:
                        pass
                del self.nodes[node]
                self.graph.remove_node(node)
            self.draw_objects()


    def add_human_graph(self):
        node_ary = [(334,274,"#BD3401"),(341,175,"#BB0001"),(405,275,"#3FBA00"),(462,354,"#00BC00"),(522,432,"#01BC3B"), (273,277,"#BB7A03"),(205,351,"#BDBB01"),(137,415,"#AAFF00"),(373,470,"#013EBF"),(376,623,"#0017AF"), (387,754,"#3502B2"),(286,465,"#00BC5F"),(278,614,"#00ACB4"),(268,753,"#026FB3"),(361,160,"#BB02AE"),(384,175,"#BB003B"),(324,159,"#7A01B7"),(294,175,"#B90278")]

        bias = len(self.graph.nodes)

        dx = random.randint(-20,20)
        dy = random.randint(-20,20)

        for x,y,color in node_ary:
            self.graph.add_node((x+dx,y+dy),color)

        edge_ary = [(0,1,"#000099"),(0,2,"#993300"),(2,3,"#669900"),(3,4,"#339900"),(0,5,"#990000"),(5,6,"#996600"),(6,7,"#999900"),(0,8,"#009999"),(8,9,"#006699"),(9,10,"#003399"),(0,11,"#009900"),(11,12,"#009933"),(12,13,"#009966"),(1,14,"#990099"),(14,15,"#990066"),(1,16,"#330099"),(16,17,"#660099")]
        for i,j,color in edge_ary:
            self.graph.add_edge(self.graph.nodes[i+bias],self.graph.nodes[j+bias],color)

        self.draw_objects()

    def move_to_front(self,event):
        self.canvas.tag_raise(self.canvas.find_withtag("current"))
        self.canvas.tag_raise("node")

    def move_to_front_human(self,event):
        obj = self.canvas.find_withtag("current")

        x,y = (None, None)
        obj_type = self.canvas.type(obj)
        if obj_type == "line" or obj_type=="polygon":
            x, y = self.edgeManager.get_coords(obj)[0:2]
        elif obj_type == "oval":
            print("oval")
            x1, y1, x2, y2 = self.canvas.coords(obj)
            x = (x1+x2)/2 
            y = (y1+y2)/2
        else:
            return

        nodes = self.get_body_nodes(self.graph.find_nearest_node(x,y))
        for node in nodes:
            for edge in node.edges:
                self.canvas.lift(self.edges[edge])
        self.canvas.tag_raise("node")


    def draw_objects(self):
        self.canvas.delete('all')
        self.nodes = {}
        self.edges = {}
        # 辺を描画する
        for edge in self.graph.edges:
            start, end = edge.start_node, edge.end_node
            x1, y1 = start.position
            x2, y2 = end.position
#            edge_id = self.canvas.create_line(x1, y1, x2, y2, width=10, fill=edge.color)
#            edge_id = create_ovalline(self.canvas, x1,y1,x2,y2, edge.color)
            edge_id = self.edgeManager.create_line(x1,y1, x2,y2, edge.color)
            print(f"edge_id={edge_id}")
            self.edges[edge] = edge_id
            self.canvas.tag_bind(edge_id, '<Button-1>', self.on_edge_pressed)
            self.canvas.tag_bind(edge_id, '<Button1-Motion>', self.drag_node_with_edge)
            self.canvas.tag_bind(edge_id, '<Button3-Motion>', self.move_nodes_with_edge)
            self.canvas.tag_bind(edge_id, '<Double-Button-3>', self.move_to_front)
            self.canvas.tag_bind(edge_id, '<Shift-Double-Button-3>', self.move_to_front_human)

        for node in self.graph.nodes:
            x, y = node.position
            node_id = self.canvas.create_oval(x-NODE_RADIUS, y-NODE_RADIUS, x+NODE_RADIUS, y+NODE_RADIUS, width=0, fill=node.color, tags="node")
            print(f"node_id={node_id}")
            self.nodes[node] = node_id
            self.canvas.tag_bind(node_id, '<Button1-Motion>', lambda event, node=node: self.drag_nodes(event, [node]))
            self.canvas.tag_bind(node_id, '<Button3-Motion>', lambda event, node=node: self.drag_nodes(event, [node]+self.get_adaptive_nodes(node)))
            self.canvas.tag_bind(node_id, '<Shift-Double-Button-3>', self.move_to_front_human)
            self.canvas.tag_bind(node_id, '<ButtonRelease-1>', self.release_node)


    def update_canvas_objects(self):
        for node in self.graph.nodes:
            canvas_node = self.nodes[node]
            coords = self.canvas.coords(canvas_node)
            cx = (coords[0] + coords[2]) / 2  # 横幅の中心座標
            cy = (coords[1] + coords[3]) / 2  # 縦幅の中心座標
            new_cx, new_cy = node.position
            new_coords = [
                coords[0] + new_cx - cx,  # 左上のx座標
                coords[1] + new_cy - cy,  # 左上のy座標
                coords[2] + new_cx - cx,  # 右下のx座標
                coords[3] + new_cy - cy   # 右下のy座標
            ]
            self.canvas.coords(canvas_node, *new_coords)  # 回転後の座標を設定

        for edge in self.graph.edges:
            start_node, end_node = edge.start_node, edge.end_node
            x1, y1 = start_node.position
            x2, y2 = end_node.position
#            self.canvas.coords(self.edges[edge], x1, y1, x2, y2)
#            coords_ovalline(self.canvas, self.edges[edge], x1, y1, x2, y2)
            self.edgeManager.set_coords(self.edges[edge], x1,y1,x2,y2)

    def get_adaptive_nodes(self,node):
        if not self.graph.nodes:
            return []
        i = self.graph.nodes.index(node)
        # 頭のノードなら頭部のノードを集めて返す
        bias = (i // 18)*18
        indices = [1+bias,14+bias,15+bias,16+bias,17+bias]
        if i in indices:
            return [self.graph.nodes[j] for j in indices]
        # それ以外なら体全部のノードを集めて返す
        else:
            return self.get_body_nodes(node)

##    save as *.json
#    def save_graph(self):
#        filename = tk.filedialog.asksaveasfilename(defaultextension='.json')
#        if filename:
#            data = {'nodes': [], 'edges': []}
#            for node in self.graph.nodes:
#                data['nodes'].append({'position': node.position, 'color': node.color})
#            for edge in self.graph.edges:
#                data['edges'].append({'start_node': self.graph.nodes.index(edge.start_node),
#                                      'end_node': self.graph.nodes.index(edge.end_node), 'color': edge.color})
#            with open(filename, 'w') as f:
#                json.dump(data, f, indent=4, separators=(',', ': '))
#                f.write('\n')

    def save_as_image(self):
        global root


        alpha = self.master.attributes("-alpha")
        self.master.attributes("-alpha", 1.0)

        filename = tk.filedialog.asksaveasfilename(defaultextension='.png')
        if filename:
            # キャンバスの座標とサイズを取得
            x = root.winfo_rootx() + self.canvas.winfo_x()
            y = root.winfo_rooty() + self.canvas.winfo_y()
            x1 = x + self.canvas.winfo_width()
            y1 = y + self.canvas.winfo_height()

            # キャンバスをキャプチャ
            # キャンバスをキャプチャ
            img = ImageGrab.grab().crop((x+2, y+2, x1-2, y1-2))
            # Graphオブジェクトを作成し、pickleを使ってバイトストリームに変換
            graph = Graph()
            # graphオブジェクトにノードやエッジを追加する処理
            graph_bytes = pickle.dumps(self.graph)

            # metadataにgraphオブジェクト情報を追加
            metadata = PngImagePlugin.PngInfo()
            metadata.add_text('graph', graph_bytes.decode('latin-1'))
            img.save(filename, pnginfo=metadata)

        self.master.attributes("-alpha", alpha)


    def load_file(self, add=False):
        # ファイル選択ダイアログを表示する
        filetypes = [('PNG and JSON files', '*.png;*.json')]
        file_path = tk.filedialog.askopenfilename(filetypes=filetypes)
        if not file_path:
            return

        # 拡張子に応じて、適切な関数を呼び出す
        if file_path.endswith('.png'):
            self.load_pose_from_image(file_path,add)
        elif file_path.endswith('.json'):
            self.load_graph(file_path,add)

    def load_pose_from_image(self, file_path=None ,add=False):
        if not file_path:
            file_path = tk.filedialog.askopenfilename(defaultextension='.png')

        # 画像を開く
        if not file_path:
            return
        with Image.open(file_path) as img:
            # metadataからgraphオブジェクト情報を取得
            graph_bytes = img.info.get('graph', None)
            if graph_bytes is not None:
                # graphオブジェクト情報があれば、バイトストリームをデシリアライズしてグラフオブジェクトを取得
                new_graph = pickle.loads(graph_bytes.encode('latin-1'))
                if add:
                    self.graph.add_graph( new_graph )
                else:
                    self.graph = new_graph
        self.draw_objects()

    def load_graph(self, filename, add=False):
#        filename = tk.filedialog.askopenfilename(defaultextension='.json')
        if not filename:
            return
        with open(filename, 'r') as f:
            graph_data = json.load(f)

        if not add:
            self.graph = Graph()
            self.nodes = {}
            self.edges = {}

        bias = len(self.graph.nodes)

        # ノードを追加する
        for node_data in graph_data['nodes']:
            position = tuple(node_data['position'])
            color = node_data['color']
            self.graph.add_node(position, color)

        # エッジを追加する
        for edge_data in graph_data['edges']:

            color = edge_data['color']
            start_node = self.graph.nodes[edge_data['start_node']+bias]
            end_node = self.graph.nodes[edge_data['end_node']+bias]
            print(f"add({start_node},{end_node})")
            self.graph.add_edge(start_node, end_node, color)

        self.draw_objects()

    def on_button3_pressed(self, event):
        self.px=event.x
        self.py=event.y

    def drag_nodes(self, event, nodes):
        x, y = event.x, event.y
        dx, dy = (x-self.px,y-self.py)
 
        nodes = list(set(nodes))

        if (event.state & 0x001):
            nodes = self.graph.nodes

        for node in nodes:
            nx, ny = (node.position[0]+dx, node.position[1]+dy)
            node.position = (nx,ny)

        self.px, self.py = x,y
        self.update_canvas_objects()

    def is_outside_canvas(self, x,y):
        return x<0 or self.canvas.winfo_width()<x or y<0 or self.canvas.winfo_height()<y

    def move_node_to(self,node,x,y):
        node.position = (x,y)
        self.update_canvas_objects()

    def on_edge_pressed(self,event):
        obj = self.canvas.find_withtag("current")
#        x1,y1,x2,y2 = self.canvas.coords(obj)
#        x1,y1,x2,y2 = endpoint_coordinates(self.canvas,obj)
        x1,y1,x2,y2 = self.edgeManager.get_coords(obj)
        

        self.inside_node = None
        self.outside_node = None
        if self.is_outside_canvas(x1,y1):
            self.outside_node = self.graph.find_nearest_node(x1,y1)
            self.inside_node = self.graph.find_nearest_node(x2,y2)
        elif self.is_outside_canvas(x2,y2):
            self.outside_node = self.graph.find_nearest_node(x2,y2)
            self.inside_node = self.graph.find_nearest_node(x1,y1)
        else:
            self.inside_node = None
            self.outside_node = None

    def drag_node_with_edge(self,event):
        if not self.outside_node or not self.inside_node:
            return 

        x,y = (event.x, event.y)
        x1,y1 = self.inside_node.position
        x2,y2 = self.outside_node.position
        l = ((x2-x1)**2 + (y2-y1)**2 )**0.5
        s = ((x-x1)**2 + (y-y1)**2 )**0.5
        new_x = x1 + l * (x-x1)/s
        new_y = y1 + l * (y-y1)/s
        self.outside_node.position = (new_x,new_y)
        
#        self.drag_nodes(event, [node])
        self.update_canvas_objects()


    def move_nodes_with_edge(self,event):
        obj = self.canvas.find_withtag("current")
        print(obj)
        x,y = (None, None)
        obj_type = self.canvas.type(obj)
        print(obj_type)
#        x, y = self.canvas.coords(obj)[0:2]
        print(obj)
        print(self.edgeManager.get_coords(obj))
        x, y = self.edgeManager.get_coords(obj)[0:2]
        node = self.graph.find_nearest_node(x,y)
        self.drag_nodes(event, [node]+self.get_adaptive_nodes(node))

    def zoom(self,event):
        x, y = self.canvas.canvasx(event.x), self.canvas.canvasy(event.y)
        scale = 1.0 + event.delta / 1200  # kを調整するためにスケールファクターを計算

        nodes = []
        if (event.state & 0x001):
            nodes = self.graph.nodes
        else:
            nodes = self.get_adaptive_nodes(self.graph.find_nearest_node(x,y))

        for node in nodes:
            x1, x2 = node.position
            node.position = ((x1-x)*scale+x, (x2-y)*scale+y)

        self.update_canvas_objects()



    def rotate(self,event):
        angle = -event.delta / 10
        x, y = self.canvas.canvasx(event.x), self.canvas.canvasy(event.y)

        nodes = []
        if (event.state & 0x001):
            nodes = self.graph.nodes
        else:
            nodes = self.get_adaptive_nodes(self.graph.find_nearest_node(x,y))

        for node in nodes:
            cx, cy = node.position
            node.position = rotate_point(cx,cy,x,y,angle)

        self.update_canvas_objects()


    def release_node(self, event):
        pass




class Graph:
    def __init__(self):
        self.nodes = []
        self.edges = []

    def add_node(self, position, color):
        node = Node(position, color)
        self.nodes.append(node)
        return node

    def add_edge(self, start_node, end_node, color):
        edge = Edge(start_node, end_node, color)
        self.edges.append(edge)
        return edge

    def add_graph(self, other):
        for node in other.nodes:
            self.nodes.append(node)
        for edge in other.edges:
            self.edges.append(edge)

    def remove_node(self, node):
        for edge in node.edges:
            try:
                self.edges.remove(edge)
            except ValueError:
                pass
        self.nodes.remove(node)

    def find_nearest_node(self, x, y):
        # 最も近い node を初期化する
        nearest_node = None
        nearest_s_distance = float('inf')

        # 各 node の座標を調べ、最も近い node を特定する
        for node in self.nodes:
            nx, ny = node.position
            s_distance = (x - nx) ** 2 + (y - ny) ** 2
            if s_distance < nearest_s_distance:
                nearest_node = node
                nearest_s_distance = s_distance

        # 最も近い node を返す
        return nearest_node

    def get_reachable_nodes(self, node):
        queue = []
        queue.append(node)

        visited = set()
        visited.add(node)

        while queue:
            node = queue.pop(0)
            neighbors = set()
            for edge in node.edges:
                neighbors.add(edge.start_node)
                neighbors.add(edge.end_node)
            for neighbor in neighbors:
                if neighbor not in visited:
                    queue.append(neighbor)
                    visited.add(neighbor)
        return list(visited)

    def edges_from_nodes(self, nodes):
        print([node.edges for node in nodes])
        edges = []
        for node in nodes:
            edges.extend(node.edges)
        return list(set(edges))

class Node:
    def __init__(self, position, color):
        self.position = position
        self.edges = []
        self.color = color

    def add_edge(self, edge):
        self.edges.append(edge)

class Edge:
    def __init__(self, start_node, end_node, color):
        self.start_node = start_node
        self.end_node = end_node
        self.color = color
        start_node.add_edge(self)
        end_node.add_edge(self)

# グラフを表示する
root = tk.Tk()
root.title("open poser 2D")
gv = GraphVisualizer(root)
gv.add_human_graph()
root.mainloop()
